/*
 * wx_cmds.h
 * Command processing
 *
 */

#ifndef cmdsh
#define cmdsh

#if (defined(HARDY_CLIPS) || defined(WXCLIPS) || defined(CLIPS5) || defined(CLIPS6) || defined(CLIPS))
#define USE_CLIPS
#endif

#ifdef USE_CLIPS
#include "clipscpp.h"
#else
#include "clipslib.h"
#endif

#ifdef GetType
#undef GetType
#endif
#define ClipsGetType(target) (target.type)

#ifdef SetType
#undef SetType
#endif
#define ClipsSetType(target, val) (target.type = (val))

// Many prototypes that used to be in wx_cmds.h, are now
// in wxextend.h. This separates out generic functionality
// from language-specific functionality (CLIP, Python, C...)
#include "wxextend.h"

#ifdef USE_CLIPS
void wxExecuteClipsFile(char *filename);
void wxUserFunctions();
#endif

// Initialize wxCLIPS
void wxInitClips(void);

// Cleanup wxClips
void wxCleanupClips(void);

// Route without echoing return result
void wxRouteNoEcho(char *buf);

#define clipsUNKNOWN 0
#define clipsSTRING  1
#define clipsDOUBLE  2
#define clipsLONG    3
#define clipsVOID    4

extern int clipsReturnType;
extern char *clipsReturnString;
extern long clipsReturnLong;
extern double clipsReturnDouble;
 
#ifdef USE_CLIPS
#define wxClipsFunction char *
#else
typedef void (*wxClipsFunction) (void);
#endif

Bool wxcObjectAddCallback(WXID id, char *eventType, wxClipsFunction callback);

/*
 * These functions are really to be compatible with C programs using
 * wxCLIPS; CLIPS doesn't need them to be defined since it has its
 * own functions and can supply the necessary arguments to the generic
 * extend functions.
 */
 
WXID wxcButtonCreate(WXID parentId, wxClipsFunction func, char *label,
  int x = -1, int y = -1, int width = -1, int height = -1, char *styleString = "");

WXID wxcButtonCreateFromBitmap(WXID parentId, wxClipsFunction func, WXID bitmapId,
  int x = -1, int y = -1, int width = -1, int height = -1, char *styleString = "");

WXID wxcTextCreate(WXID parentId, wxClipsFunction func, char *label, char *value,
  int x, int y, int width, int height, char *styleString);

WXID wxcMultiTextCreate(WXID parentId, wxClipsFunction func, char *label, char *value,
  int x, int y, int width, int height, char *styleString);

WXID wxcCheckBoxCreate(WXID parentId, wxClipsFunction func, char *label,
  int x = -1, int y = -1, int width = -1, int height = -1, char *styleString = "");

WXID wxcSliderCreate(WXID parentId, wxClipsFunction func, char *label,
  int value, int min_value, int max_value, int width, int x = -1, int y = -1, char *styleString = "");

WXID wxcListBoxCreate(WXID parentId, wxClipsFunction func, char *label, Bool multiple = FALSE,
  int x = -1, int y = -1, int width = -1, int height = -1, char *styleString = "");

WXID wxcChoiceCreate(WXID parentId, wxClipsFunction func, char *label, int x, int y,
  int width, int height, int n, char **choices, char *styleString);

WXID wxcRadioBoxCreate(WXID parentId, wxClipsFunction func, char *label, int x, int y,
  int width, int height, int n, char **choices, int majorDim, char *styleString);

#ifndef USE_CLIPS
WXID wxcChoiceCreate(WXID parentId, wxClipsFunction func, char *label, int x = -1, int y = -1,
  int width = -1, int height = -1);

long wxcChoiceCreate(long parent_id, wxClipsFunction func, char *label, long x, long y,
  long width, long height, ClipsExpr& choices, char *style = "");

long wxcRadioBoxCreate(long parent_id, wxClipsFunction func, char *label, long x, long y,
  long width, long height, ClipsExpr& choices, int majorDim, char *style = "");

char *wxcGetChoice(char *message, ClipsExpr& choices, long centre = TRUE, long parentId = 0);
#endif

#endif // cmdsh
